
/*
 * check_loc.c
 * Copyright (C) 1997 by USC/ISI
 * $Id: check_loc.c,v 1.1 2000/04/13 04:31:53 johnh Exp $
 *
 * Copyright (c) 1997 University of Southern California.
 * All rights reserved.                                            
 *                                                                
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation, advertising
 * materials, and other materials related to such distribution and use
 * acknowledge that the software was developed by the University of
 * Southern California, Information Sciences Institute.  The name of the
 * University may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include "cron.h"
#include "defs.h"
#include <stdio.h>
#include <string.h>
#include <errno.h>


char * progname;		/* Name we are running under */

extern void setlibdir();
extern int sendcommand();

int 
main(argc, argv)
	int argc;
	char *argv[];
{
	FILE *f;
#define LINELEN 1024
	char line[LINELEN];
	char prog[LINELEN];
	char *netstat;
	char *token_temp;
	char *token_location;
	char *token_flag;
	int i;
	static char *netstats[] = {
		"/usr/bin/netstat",
		"/bin/netstat",
		"/etc/netstat",
		"/usr/ucb/netstat",
		NULL
	};
	struct stat sb;
	int re_matches,u_flag_ok,ug_flag_ok;

	/* find netstat */
	for (i = 0; netstats[i]; i++) {
		netstat = netstats[i];
		if (stat(netstat, &sb) == 0)
			break;
	};
	if (netstat == NULL) {
		fprintf(stderr, "%s: cannot find netstat in compiled-in path.\n");
		exit(1);
	};

	progname = argv[0];
	setlibdir();
	if (-1 == chdir("events")) {
		fprintf(stderr, "%s: cannot chdir to LIBDIR/entries\n");
		exit(1);
	};

	if (NULL == (outf = fopen("gateways", "w"))) {
		fprintf(stderr, "%s: cannot create gateways\n");
		exit(1);
	};

	/*
	 * Run and parse netstat, saving the output.
	 */
	sprintf(prog, "%s -rn", netstat);
	if (NULL == (inf = popen(prog, "r"))) {
		fprintf(stderr, "%s: cannot run netstat\n");
		exit(1);
	};
	while (fgets(line, LINELEN, inf)) {
		if (NULL == (token_temp = strtok(line," ")))
			continue;
		if (NULL == (token_location = strtok(NULL," ")))
			continue;
		if (NULL == (token_flag = strtok(NULL," ")))
			continue;
		if (isdigit(token_flag[0]))   /* in Linux, skip over genmask field */
			if (NULL == (token_flag = (char*)strtok(NULL," ")))
				continue;
		/*
		 * At last, may have a gateway.
		 */
		re_matches = re_exec(token_location);
		u_flag_ok = !strcmp(token_flag, "U");
		ug_flag_ok = !strcmp(token_flag, "UG");
		if (re_matches && (u_flag_ok || ug_flag_ok)) {
			fprintf(outf, "%s\n", token_location);
		};

	};
	close(outf);
	pclose(inf);

	/*
	 * Signal cron.
	 */
	switch (sendcommand('L', "", 1)) {
	case 0: break;
	case 1: fprintf(stderr, "%s: error: cron not running\n", progname); break;
	default: fprintf(stderr, "%s: error: some problem signalling cron\n", progname); break;
	};
}
