#ifndef lint
static char Rcs_Id[] =
    "$Header: /h/guests/fmg/johnh/GEOFF_CVS/xcron/limits.c,v 2.15 1998/07/30 20:42:25 geoff Exp $";
#endif

/*
 * $Log: limits.c,v $
 * Revision 2.15  1998/07/30 20:42:25  geoff
 * Add support for resource-exhaustion limits.  Fix a rather nasty bug in
 * the way defaults were processed, that could cause most defaults to be
 * ignored.
 *
 * Revision 2.14  1995/11/27  06:13:58  geoff
 * Fix bugs in the last delta which removed strtok
 *
 * Revision 2.13  1995/11/27  05:03:52  geoff
 * Don't use strtok to parse userlimits, at the expense of a more
 * restrictive file format.
 *
 * Revision 2.12  1995/11/27  01:59:39  geoff
 * Include errno.h for the debugging macros.
 *
 * Revision 2.11  1995/05/11  05:05:29  geoff
 * Minor RCS header changes.
 *
 * Revision 2.10  87/03/31  22:32:56  geoff
 * Make the include of string.h configurable.
 * 
 * Revision 2.9  87/01/10  19:19:45  geoff
 * Set defaults before opening the userlimits file, in case it doesn't
 * exist.
 * 
 * Revision 2.8  86/11/29  19:51:50  geoff
 * Change all the register declarations to be portable across
 * machines with varying numbers of registers.
 * 
 * Revision 2.7  86/04/17  21:55:45  geoff
 * Change the "limits read" log message to support the crontab -L switch.
 * 
 * Revision 2.6  86/04/01  19:02:48  geoff
 * Fix a preprocessor argument-mismatch bug introduced in the last revision.
 * 
 * Revision 2.5  86/03/31  22:58:13  geoff
 * Add queue length limits, and fix a bug in an error message.
 * 
 * Revision 2.4  86/03/30  15:41:20  geoff
 * Switch debugging output to the DBPRINTn macros when possible
 * 
 * Revision 2.3  86/03/30  14:33:36  geoff
 * Add the setdefaults routine and don't initialize userdefaults.  Put out
 * a log message when the limits file is read.  Add code to reset all user
 * defaults when a default line is read.
 * 
 * Revision 2.2  86/03/28  14:41:51  geoff
 * Fix some lint complaints and minor bugs, and add NEVERDEL support.
 * 
 * Revision 2.1  86/03/26  00:21:41  geoff
 * Initial Revision.
 * 
 */

#include "cron.h"
#include <ctype.h>
#include <errno.h>
#include <pwd.h>
#include <stdio.h>
#include "pstring.h"
#include <time.h>

void		readlimits ();	/* Read in the limits file */
static void	userlimits ();	/* Read limits for a user */
static void	syntax ();	/* Issue a syntax error message */
static void	setdefaults ();	/* Initialize limits to defaults */

struct cronuser	userdefaults;	/* Default limits for users */

static int	lineno = 1;	/* Current line in file */

extern struct cronuser *
		adduser ();	/* Add (or find) a cron user entry */
extern struct passwd *
		getpwnam ();	/* Get password entry by name */
extern struct passwd *
		getpwuid ();	/* Get password entry by UID */
extern time_t	time ();	/* Return system time */
extern char *	timenonl ();	/* Return ASCII time without a newline */
extern struct cronuser
		userhead;	/* Head of cron user list */

void readlimits ()			/* Read in the limits file */
					/* ..must be called from LIBDIR */
    {
    char		buf[BUFSIZ];	/* Buffer for lines from file */
    reg1 FILE *		limfile;	/* Access to the limits file */
    int			lineinc;	/* Increment for line number */
    time_t		rtime;		/* Time file was read, for log */
    reg2 char *		sep;		/* Pointer to field separator */

    DBPRINT0 (12, "readlimits ()\n");
    setdefaults ();
    limfile = fopen (LIMITFILE, "r");
    if (limfile == NULL)
	{
	DBPRINT2 (1, "readlimits:  couldn't open %s/%s\n", LIBDIR, LIMITFILE);
	return;
	}
    (void) fseek (stderr, 0L, 2);
    rtime = time ((time_t *) NULL);
    (void) fprintf (stderr, "! *** %s/%s read %s\n", LIBDIR, LIMITFILE,
      timenonl (rtime));
    for (  ;  fgets (buf, sizeof buf, limfile) != NULL;  lineno += lineinc)
	{
	if (strchr (buf, '\n') != NULL)
	    lineinc = 1;
	else
	    lineinc = 0;
	if ((sep = strchr (buf, ' ')) == NULL)
	    continue;
	*sep++ = '\0';
	if (strcmp (buf, "user") == 0)
	    userlimits (sep);
	else
	    syntax ();
	}
    (void) fclose (limfile);
    }

static void userlimits (buf)		/* Read limits for a user */
    reg1 char *		buf;		/* Buffer being parsed */
    {
    reg4 char *		fieldname;	/* Field being changed */
    reg6 char *		fieldval;	/* Value of the field */
    struct passwd *	pwent;		/* User's password file entry */
    reg2 char *		sep;		/* Pionter to next separator */
    reg3 struct cronuser * userent;	/* Descriptive entry for the user */
    reg5 char *		username;	/* Name of the user */

    sep = strchr (buf, '\n');
    if (sep != NULL)
	*sep = '\0';
    sep = strchr (buf, ' ');
    username = buf;
    if (sep == NULL)
	{
	syntax ();
	return;
	}
    *sep++ = '\0';
    buf = sep;
    if (strcmp (username, "default") == 0)
	userent = &userdefaults;
    else
	{
	if (isdigit (username[0]))
	    pwent = getpwuid (atoi (username));
	else
	    pwent = getpwnam (username);
	if (pwent == NULL)
	    {
	    (void) fseek (stderr, 0L, 2);
	    (void) fprintf (stderr,
	      "readlimits:  uid '%s' on line %d is unrecognized\n",
	      username, lineno);
	    return;
	    }
	userent = adduser (pwent->pw_uid);
	if (userent == NULL)
	    return;
	}
    DBPRINT1 (14, "userlimits: doing user %s\n", username);
    while (*buf != '\0')
	{
	sep = strchr (buf, ' ');
	if (sep == NULL)
	    {
	    syntax ();
	    return;
	    }
	*sep++ = '\0';
	fieldname = buf;
	fieldval = sep;
	sep = strchr (sep, ' ');
	if (sep == NULL)
	    buf = "";
	else
	    {
	    *sep++ = '\0';
	    buf = sep;
	    }
	DBPRINT2 (15, "userlimits: doing field %s = %s\n",
	  fieldname, fieldval);
	if (strcmp (fieldname, "procs") == 0)
	    userent->cu_proclim = atoi (fieldval);
	else if (strcmp (fieldname, "queue") == 0)
	    userent->cu_qlim = atoi (fieldval);
	else if (strcmp (fieldname, "nice") == 0)
	    userent->cu_nice = atoi (fieldval);
	else if (strcmp (fieldname, "entries") == 0)
	    userent->cu_entlim = atoi (fieldval);
	else if (strcmp (fieldname, "entsize") == 0)
	    userent->cu_entsizelim = atol (fieldval);
	else if (strcmp (fieldname, "mailsize") == 0)
	    userent->cu_mailsizelim = atol (fieldval);
	else
	    {
	    syntax ();
	    return;
	    }
	}
    if (userent == &userdefaults)
	{
	/*
	 * A default line has been read.  Go through the users and
	 * reset their default values.
	 */
	for (userent = userhead.cu_next;
	  userent != &userhead;
	  userent = userent->cu_next)
	    {
	    userent->cu_flags &= ~UF_NEVERDEL;
	    userent->cu_proclim = userdefaults.cu_proclim;
	    userent->cu_qlim = userdefaults.cu_qlim;
	    userent->cu_nice = userdefaults.cu_nice;
	    userent->cu_entlim = userdefaults.cu_entlim;
	    userent->cu_entsizelim = userdefaults.cu_entsizelim;
	    userent->cu_mailsizelim = userdefaults.cu_mailsizelim;
	    }
	}
    else
	{
	userent->cu_flags &= ~UF_MARKDEL;	/* Never delete */
	userent->cu_flags |= UF_NEVERDEL;
	}
    }

static void syntax ()			/* Issue syntax error message */
    {
    (void) fseek (stderr, 0L, 2);
    (void) fprintf (stderr, "readlimits:  syntax error on line %d\n",
      lineno);
    }

static void setdefaults ()		/* Set limits to their defaults */
    {
    reg1 struct cronuser *	userp;	/* Current user structure */

    userdefaults.cu_proclim = DEFPROCLIM;
    userdefaults.cu_qlim = DEFQLIM;
    userdefaults.cu_nice = DEFNICE;
    userdefaults.cu_entlim = DEFENTLIM;
    userdefaults.cu_entsizelim = DEFENTSIZELIM;
    userdefaults.cu_mailsizelim = DEFMAILSIZELIM;
    for (userp = userhead.cu_next;
      userp != &userhead;
      userp = userp->cu_next)
	{
	userp->cu_flags &= ~UF_NEVERDEL;
	userp->cu_proclim = DEFPROCLIM;
	userp->cu_qlim = DEFQLIM;
	userp->cu_nice = DEFNICE;
	userp->cu_entlim = DEFENTLIM;
	userp->cu_entsizelim = DEFENTSIZELIM;
	userp->cu_mailsizelim = DEFMAILSIZELIM;
	}
    }
