#ifndef lint
static char Rcs_Id[] =
    "$Header: /h/guests/fmg/johnh/GEOFF_CVS/xcron/timesubs.c,v 2.10 2000/04/14 01:17:54 johnh Exp $";
#endif

/*
 * $Log: timesubs.c,v $
 * Revision 2.10  2000/04/14 01:17:54  johnh
 * patch for y2k problem in at jobs
 *
 * Revision 2.9  2000/04/08 00:15:36  geoff
 * Never generate a negative weekday.
 *
 * Revision 2.8  1995/11/27 01:59:54  geoff
 * Support mktime() if the system has it.
 *
 * Revision 2.7  1995/05/11  05:06:01  geoff
 * Switch to DST on the first Sunday of April, not the last.
 *
 * Revision 2.6  88/10/31  23:54:01  geoff
 * Get rid of the complexities of election-year DST tables.  In the first
 * place, they seem to have been repealed by our ever-wise Congress.  In
 * the second, cron will either hang or spawn infinite processes if these
 * tables don't agree with those used by localtime(), which I don't have
 * control over.
 * 
 * Revision 2.5  87/04/01  23:51:21  geoff
 * Add U.S. daylight-savings tables good for up to 1999.
 * 
 * Revision 2.4  86/11/29  19:54:51  geoff
 * Change all the register declarations to be portable across
 * machines with varying numbers of registers.
 * 
 * Revision 2.3  86/11/16  13:18:17  geoff
 * Correct the 1987 DST table to be what Congress finally chose
 * 
 * Revision 2.2  86/04/13  13:33:24  geoff
 * Add the second parameter to dstamount, and fix some bugs in it.
 * 
 * Revision 2.1  86/03/28  13:42:29  geoff
 * Initial Revision.
 * 
 */

#include "config.h"
#include <stdio.h>
#include <time.h>
#include <sys/types.h>

time_t		tm_to_time ();	/* Convert struct tm to time_t */
time_t		dstamount ();	/* Calculate daylight savings offset */
int		wkday ();	/* Calculate day of the week */

#ifndef HASMKTIME
/*
 * Daylight-savings description table */
struct dsttab
    {
    short		syear;		/* First year conversion applies */
    short		eyear;		/* Last year it applies in */
    short		smonth;		/* First month it applies in */
					/* ZERO-INDEXED! */
    short		sday;		/* First day it applies on */
    short		scode;		/* Code for forcing day of week */
					/* -1 == use date given */
					/* 0-6 == search fwd to that weekday */
    time_t		stime;		/* When to start doing daylight */
    short		emonth;		/* Last month it applies in */
					/* ZERO-INDEXED! */
    short		eday;		/* Last day it applies on */
    short		ecode;		/* Code for forcing day of week */
    time_t		etime;		/* When daylight time ends */
    time_t		offset;		/* Offset to apply during period */
					/* ..This is assumed to be positive */
    };

/*
 * BUG:  This table should be malloc'ed and read in from a file or
 * environment variable, rather than being compiled in.  Unfortunately,
 * the people involved haven't figured out a standard yet.
 *
 * BUG:  The entries in this table *must* agree with the table used by the
 * localtime routine.  Otherwise, cron may get into a loop spawning endless
 * jobs, sigh.
 */
static struct dsttab	dstimes[] = /* Table of dst conversions */
    {
    /***** WARNING:  Remember the month starts at zero! *****/
  /* sy   ey sm[0]  sd sc   stime       em[0]  ed ec       etime   offset */ 
		    /* U.S. */
    { 0,  99,    3,  1, 0, 2 * 60 * 60,     9, 25, 0, 2 * 60 * 60, 60 * 60},
  /* sy   ey sm[0]  sd sc   stime       em[0]  ed ec       etime   offset */ 
    /***** WARNING:  Remember the month starts at zero! *****/
    };
#endif /* HASMKTIME */

static short	tz_is_set = 0;	/* NZ once tzset() has been called */

/*
 * Leap-year calculation.  This will work for all dates in the future until
 * the year 2100, by which time I hope this software is thoroughly obsolete.
 * Also note that it works whether applied to a 4-digit year or a Unix
 * "year - 1900".
 */
#define leap(year)	(((year) & 3) == 0)

/*
 * Leap-year adjustment calculation.  Generates 1 if month is January or
 * February and the year is a leap year.  Note months are indexed from zero.
 *
 * Because at.c calls this with a 0-based date, 
 * we check and handle both 0-based and 1900-based dates. -johnh, 13-Apr-00
 */
#define leapadj(month, year)	((month) <= 1  &&  leap (year))

extern int	daylight;	/* NZ if daylight savings time applies */
extern long	timezone;	/* System timezone in seconds */

time_t tm_to_time (buf)			/* Convert struct tm to time_t */
    reg1 struct tm *	buf;		/* Buffer to convert */
    {
#ifdef HASMKTIME
    int year_correction = 0;
    time_t result;

    buf->tm_isdst = -1;
    if (buf->tm_year >= 1900) {
	    year_correction = -1900;
	    buf->tm_year += year_correction;
    }
    result = mktime(buf);
    if (year_correction)
	    buf->tm_year -= year_correction;
    return result;
#else /* HASMKTIME
    /*
     * Table of the number of days between the beginning of the year
     * and the first day of a month.
     */
    static short		montab[] =
				    {
				    0,
				    31,
				    31+28,
				    31+28+31,
				    31+28+31+30,
				    31+28+31+30+31,
				    31+28+31+30+31+30,
				    31+28+31+30+31+30+31,
				    31+28+31+30+31+30+31+31,
				    31+28+31+30+31+30+31+31+30,
				    31+28+31+30+31+30+31+31+30+31,
				    31+28+31+30+31+30+31+31+30+31+30
				  /* j  f  m  a  m  j  j  a  s  o  n */
				    };
    reg2 int			leaps;	/* Count of leap days */
    reg3 time_t			rtime;	/* Return time value */
    int year_from_1900;

    if (!tz_is_set)
	{
	tzset ();
	tz_is_set = 1;
	}
    /* correct for internal years */
    year_from_1900 = buf->tm_year;
    if (year_from_1900 >= 1900)
	    year_from_1900 -= 1900;
    rtime = (year_from_1900 - 70) * 365;
    /*
     * Calculate the number of leap days between 1970 and now.
     */
    leaps = (year_from_1900 - 70 + 2) / 4; /* Includes this year if it's leap */
    leaps -= leapadj (buf->tm_mon, year_from_1900); /* Correct that if Jan/Feb */
    rtime += leaps + montab[buf->tm_mon] + buf->tm_mday - 1;
    rtime *= 24;			/* Convert from days to hours */
    rtime += buf->tm_hour;
    rtime *= 60;
    rtime += buf->tm_min;
    rtime *= 60;
    rtime += buf->tm_sec;
    return rtime - dstamount (buf,  0) + timezone;
#endif /* HASMKTIME */
    }

time_t dstamount (when, dir)		/* Calculate daylight savings offset */
    reg2 struct tm *		when;	/* Buffer with broken-down time */
    reg5 int			dir;	/* NZ if going to, not from, DST */
    {
#ifdef HASMKTIME
    time_t			time_with_dst; /* Converted time, with DST */
    time_t			time_without_dst; /* Converted time, no DST */

    when->tm_isdst = 0;
    time_without_dst = mktime(when);
    when->tm_isdst = dir ? -1 : 1;
    time_with_dst = mktime(when);
    return time_with_dst - time_without_dst;
#else /* HASMKTIME */
    reg1 struct dsttab *	dstptr;	/* Pointer into dsttab */
    reg3 int			eday;	/* Actual end day this year */
    reg4 int			sday;	/* Actual start day this year */

    if (!tz_is_set)
	{
	tzset ();
	tz_is_set = 1;
	}
    if (!daylight)
	return 0;			/* Daylight conversion disabled */
    for (dstptr = dstimes;
      dstptr < &dstimes[sizeof dstimes / sizeof dstimes[0]];
      dstptr++)
	{
	if (when->tm_year >= dstptr->syear  &&  when->tm_year <= dstptr->eyear
	  &&  when->tm_mon >= dstptr->smonth
	  &&  when->tm_mon <= dstptr->emonth)
	    {				/* Right year, right month range */
	    if (when->tm_mon == dstptr->smonth)
		{			/* Check exact day */
		if (dstptr->scode >= 0)
		    {
		    if (when->tm_mday < dstptr->sday)
			continue;	/* Avoid running wkday if we can */
		    sday = wkday (when->tm_year, when->tm_mon, dstptr->sday);
		    sday = dstptr->scode - sday;
		    if (sday < 0)
			sday += 7;
		    sday += dstptr->sday;
		    }
		else
		    sday = dstptr->sday;
		if (when->tm_mday < sday)
		    continue;
		else if (when->tm_mday == sday)
		    {			/* It's the day daylight starts! */
		    /*
		     * We may be into the daylight-savings range.  However,
		     * because there is a discontinuity, it is possible
		     * that "when" specifies a nonexistent time during
		     * the discontinuity.  If this is true, the best way
		     * to treat this depends on the direction of conversion.
		     * If 'dir' is zero (we are converting from
		     * possibly-daylight to standard) we should
		     * still treat it as standard time;  this takes the
		     * nonexistent 2:30 a.m. on spring daylight day and
		     * turns it into 3:30 am daylight time.  If 'dir' is
		     * nonzero (we are converting from standard to
		     * possibly-daylight) there is never a discontinuity
		     * and we want to arrange for 1:59 am to be standard,
		     * but 2:00 to be daylight.  The ?: expression below
		     * accomplishes this.
		     */
		    if (((when->tm_hour * 60) + when->tm_min) * 60
			+ when->tm_sec
		      < dstptr->stime + (dir ? 0 : dstptr->offset))
			continue;	/* But not time to start yet */
		    }
		}
	    else if (when->tm_mon == dstptr->emonth)
		{			/* Check exact day */
		if (dstptr->ecode >= 0)
		    {
		    if (when->tm_mday > dstptr->eday + 6)
			continue;	/* Avoid running wkday if we can */
		    eday = wkday (when->tm_year, when->tm_mon, dstptr->eday);
		    eday = dstptr->ecode - eday;
		    if (eday < 0)
			eday += 7;
		    eday += dstptr->eday;
		    }
		else
		    eday = dstptr->eday;
		if (when->tm_mday > eday)
		    continue;
		else if (when->tm_mday == eday)
		    {			/* It's the day daylight ends! */
		    if (((when->tm_hour * 60) + when->tm_min) * 60
			+ when->tm_sec
		      >= dstptr->etime)
			continue;	/* But past the time it ends */
		    }
		}
	    /*
	     * The only way to get this point is if daylight savings
	     * time is in effect for the date that was passed to us
	     * as an argument.
	     */
	    return dstptr->offset;
	    }
	}
    return 0;				/* Must not be on dst */
#endif HASMKTIME
    }

int wkday (year, month, day)		/* Calculate day of the week */
    reg1 int			year;	/* Year - 1900 */
    reg2 int			month;	/* Month, zero-indexed */
    int				day;	/* Day of month, one-indexed */
    {
    static short adj[12] = {1, 4, 4, 0, 2, 5, 0, 3, 6, 1, 4, 6};
    reg3 int			wday;	/* Space for the result */

    /* This is a formula I learned from the net. */
    wday =
      year / -50			/* CC */
      + ((year % 100) * 5) / 4		/* + 1.25 * YY */
      + adj[month]			/* + months[mon] */
      + day				/* + dd */
      - leapadj (month, year);		/* - fudge */
    wday %= 7;				/* mod 7, forcing result positive */
    if (wday < 0)
	wday += 7;
    return wday;
    }
